#!/usr/bin/env python
# -*- coding: utf-8 -*-

from gaidaros import Gaidaros

def multiline_split(x):
    _loc = x.find('\n\n') + 2
    if _loc - 2 < len(x):
        return (x[:_loc], x[_loc:])
    else:
        return (x, '')

server = Gaidaros(end_request = lambda x: '\n\n' in x, split_request = multiline_split)
server.serve()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import with_statement
import socket as sock, ssl, multiprocessing as mp, time, subprocess as sp, tempfile as tf, re
from gaidaros import Gaidaros

def tests():
    def _do_single(*args, **kwargs):
        srv = Gaidaros(*args, **kwargs)
        srv.handle()

    def _do_multiple(*args, **kwargs):
        kwargs.update({'handle_request': lambda x: (x, True)})
        srv = Gaidaros(*args, **kwargs)
        srv.serve()

    _patt = re.compile(r'@HostName@')
    hostname ='localhost'
    with tf.NamedTemporaryFile(suffix='.pem',prefix='gaidaros.') as tempcert:
        with tf.NamedTemporaryFile(suffix='.cnf',prefix='gaidaros.') as tempconf:
            with open('/usr/share/ssl-cert/ssleay.cnf', 'r') as conftemplate:
                for _line in conftemplate.readlines():
                    tempconf.write(_patt.sub('"' + hostname + '"', _line))
                tempconf.flush()
                sp.check_call(['openssl', 'req', '-config', tempconf.name, '-new', '-x509', '-days', '1',
                               '-nodes', '-out', tempcert.name, '-keyout', tempcert.name])
                sp.check_call(['chmod', '644', tempcert.name]) #throwaway cert,no privacy-concern
        _version = (sock.AF_INET, sock.AF_INET6)
        _type = ('single', 'multiple')
        _security = ('plain', 'ssl')
        _sslversion = (ssl.PROTOCOL_TLSv1, ssl.PROTOCOL_SSLv23, ssl.PROTOCOL_SSLv3)
        _role = ('server', 'client')

        for _v in _version:
            for _t in _type:
                for _s in _security:
                    for _sv in _sslversion:
                        for _r in _role:
                            if _r == 'server':
                                _srv_kwargs = {'port': 4000}
                                if _s == 'ssl':
                                    _srv_kwargs.update(
                                        {'use_ssl': True, 'ssl_certfile': tempcert.name, 'ssl_version': _sv})
                                if _t == 'single':
                                    _target=_do_single
                                else:
                                    _target=_do_multiple
                                _srvsock = mp.Process(target=_target, kwargs=_srv_kwargs)
                                _srvsock.daemon = True
                                _srvsock.start()
                                time.sleep(1)
                            else:
                                _clientsock = sock.socket(_v, sock.SOCK_STREAM)
                                _clientsock.connect((hostname, 4000))
                                _clientsock.setsockopt(sock.SOL_SOCKET, sock.SO_REUSEADDR, 1)
                                if _s == 'ssl':
                                    _clientsock = ssl.wrap_socket(_clientsock, cert_reqs=ssl.CERT_NONE, ssl_version=_sv)
                                if _t == 'single':
                                    _reps = 1
                                else:
                                    _reps = 1024
                                _response = ''
                                for _count in xrange(_reps):
                                    _clientsock.sendall('abc-αβγ-123\n')
                                    _response += _clientsock.recv(1024)
                                _clientsock.close()
                                _srvsock.terminate()
                                assert _response == 'abc-αβγ-123\n' * _reps



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from gaidaros import Gaidaros

# ...handles Unix and MS EOLs

server = Gaidaros(handle_request = lambda x: ("Hello World: you said \"{}\"\r\n".format(x.rstrip('\r\n')), False)
server.handle()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#NB: "unicode" introduces more problems than it solves, if we want to also be able to
#    handle arbitrary binary data "strings" (which often don't have a unicode
#    representation). For that reason we now only use binstrings, and the handler has
#    to deal with it explicitly if you want it. Even if you override decode_request()
#    and encode_response() to decode to Python's "unicode" in the handler, don't be
#    tempted to pass back u'' instead of b'', it will throw a TypeError. Do the
#    b'' -> u'' inside handle_request() (and maybe end_request() and split_request())
#    if you really need it, but it *will* slow things down. My tests instantly started
#    serving five times faster when I removed the redundant UTF-8 -> unicode() logic...

"""
Async server micro-framework for control freaks
"""
__version__ = '0.3.13'

import sys, os, ConfigParser, inspect, importlib, socket, select, errno, ast
try:
    import ssl
    has_ssl = True
except ImportError:
    has_ssl = False

def _print_nl(data, ostream=sys.stdout):
    ostream.write(data + ('\n', '')[data[-1:] == '\n'])

def usage(ostream=sys.stderr):
    pass

def warn(warning, ostream=sys.stderr, with_usage=False):
    if with_usage:
        usage(ostream=ostream)
        ostream.write('\n')
    _print_nl(os.path.basename(__file__) + ': ' + warning, ostream=ostream)

def die(warning=None, stack=None, with_usage=False):
    if warning:
        warn(warning, with_usage=with_usage)
    if __name__ == '__main__':
        if stack:
            sys.stderr.write('\n' + stack)
        sys.exit(1)
    raise

def log(*args):
    for message in args:
        warn(message, ostream=sys.stderr)

class Gaidaros(object):

    _globals = globals()
    
    def __init__(self, verbose=None, conf=None, host=None, port=None, ip_version=None, \
                 backlog=None, poll_timeout=None, recv_size=None, \
                 use_ssl=None, ssl_certfile=None, ssl_keyfile=None, \
                 ssl_cert_reqs=None, ssl_ca_certs=None, ssl_version=None, die_on_error=None, \
                 handler_class=None, handler_class_args=None, handler_class_kwargs=None, handler_module=None, \
                 handle_request=None, end_request=None, split_request=None, \
                 decode_request=None, encode_response=None):
        _proc = {
            'end_request': end_request,
            'split_request': split_request,
            'handle_request': handle_request,
            'decode_request': decode_request,
            'encode_response': encode_response,
        }
        for _prockey in _proc.keys(): # eager-evaluate as we are changing the dict in the loop
            _proc[_prockey + '_name'] = None
        if conf:
            conf = str(conf)
            ## source conf for default settings
            self.cnf = ConfigParser.ConfigParser()
            self.cnf.MAX_INTERPOLATION_DEPTH = 3
            cnf_fp = open(os.path.expanduser(conf))
            self.cnf.readfp(cnf_fp)
            cnf_fp.close()
            sys_path = self.cnf.get('dir', 'lib')
            sys_run = self.cnf.get('dir', 'run')
            if sys_path and sys_path not in sys.path:
                sys.path.insert(0, os.path.expanduser(sys_path))
            if verbose is None:
                verbose = self.cnf.getboolean('parameter', 'verbose')
            if host is None:
                host = self.cnf.get('parameter', 'host')
            if port is None:
                port = self.cnf.getint('parameter', 'port')
            if ip_version is None:
                ip_version = self.cnf.getint('parameter', 'ip_version')
            if backlog is None:
                backlog = self.cnf.getint('parameter', 'backlog')
            if poll_timeout is None:
                poll_timeout = self.cnf.getint('parameter', 'poll_timeout')
            if recv_size is None:
                recv_size = self.cnf.getint('parameter', 'recv_size')
            if use_ssl is None:
                use_ssl = self.cnf.getboolean('parameter', 'use_ssl')
            if ssl_certfile is None:
                ssl_certfile = self.cnf.get('parameter', 'ssl_certfile')
            if ssl_keyfile is None:
                ssl_keyfile = self.cnf.get('parameter', 'ssl_keyfile')
            if ssl_cert_reqs is None:
                ssl_cert_reqs = self.cnf.getint('parameter', 'ssl_cert_reqs')
            if ssl_ca_certs is None:
                ssl_ca_certs = self.cnf.get('parameter', 'ssl_ca_certs')
            if ssl_version is None:
                ssl_version = self.cnf.getint('parameter', 'ssl_version')
            if die_on_error is None:
                die_on_error = self.cnf.getboolean('parameter', 'die_on_error')
            if handler_module is None:
                handler_module = self.cnf.get('handler', 'module')
            if handler_class is None:
                handler_class = self.cnf.get('handler', 'class')
            if handler_class_args is None:
                _tmpvar = self.cnf.get('handler', 'class_args')
                if _tmpvar not in (None, ''):
                    handler_class_args = tuple(ast.literal_eval(_tmpvar))
            if handler_class_kwargs is None:
                _tmpvar = self.cnf.get('handler', 'class_kwargs')
                if _tmpvar not in (None, ''):
                    handler_class_kwargs = dict(ast.literal_eval(_tmpvar))
            for _prockey in _proc:
                if _prockey[-5:] != '_name' and _proc[_prockey] in (None, ''):
                    _proc[_prockey] = self.cnf.get('handler', _prockey)
        ## set the defaults not specified in the conf
        if verbose in (None, ''):
            verbose = False
        if port in (None, ''):
            port = 8080
        if ip_version in (None, ''):
            ip_version = 0
        if host in (None, ''):
            if ip_version in (0, 6):
                host = '::'
            else:
                host = '0.0.0.0'
        if backlog in (None, ''):
            backlog = 50
        if poll_timeout in (None, ''):
            poll_timeout = 1
        if recv_size in (None, ''):
            recv_size = 1024
        if use_ssl in (None, ''):
            use_ssl = False
        if ssl_cert_reqs in (None, ''):
            ssl_cert_reqs = ssl.CERT_NONE
        if ssl_ca_certs in (None, ''):
            ssl_ca_certs = '/etc/ca-certificates.conf' # Debian default location
        if ssl_version in (None, ''):
            ssl_version = ssl.PROTOCOL_TLSv1
        if die_on_error in (None, ''):
            die_on_error = True
        if handler_module in (None, ''):
            handler_module_name = ''
            handler_module = None
        elif isinstance(handler_module, basestring):
            handler_module_name = handler_module
            handler_module = importlib.import_module(handler_module_name)
        elif inspect.ismodule(handler_module):
            if handler_module_name in (None, ''):
                handler_module_name = handler_module.__name__
        else:
            raise TypeError('handler_module is not a valid module or name of one. It is: {} (type {})'.format(handler_module, type(handler_module)))
        if handler_class in (None, ''):
            handler_class_name = ''
            handler_class = None
            handler_class_args = ()
            handler_class_kwargs = {}
        elif isinstance(handler_class, basestring):
            handler_class_name = str(handler_class)
            if handler_module is not None and hasattr(handler_module, handler_class_name) and inspect.isclass(getattr(handler_module, handler_class_name)):
                handler_class = getattr(handler_module, handler_class_name)
            elif self._globals.has_key(handler_class_name) and inspect.isclass(eval(handler_class_name)):
                handler_class = eval(handler_class_name)
            else:
                raise ValueError('handler_class is not the name of a valid class. It is: {}'.format(handler_class_name))
        elif inspect.isclass(handler_class):
            if handler_class_name in (None, ''):
                handler_class_name = str(handler_class.__name__)
        else:
            raise TypeError('handler_class is not a valid class. It is: {} (type {})'.format(handler_class, type(handler_class)))
        for _prockey in _proc:
            if _prockey[-5:] != '_name' and _proc[_prockey] in (None, ''):
                _proc[_prockey] = _prockey
        if use_ssl and not has_ssl:
            raise NotImplementedError('use_ssl specified but ssl module not available or usable')
        ## set long-lived vars including handler instance
        for x in 'verbose', 'host', 'port', 'ip_version', 'backlog', 'poll_timeout', \
                 'recv_size', 'use_ssl', 'ssl_certfile', 'ssl_keyfile', 'ssl_cert_reqs', \
                 'ssl_ca_certs', 'ssl_version', 'die_on_error':
            setattr(self, x, locals()[x])
        # POSIX.1-2001 allows either error, and doesn't require them to have same value,
        # so check for both (merging identical integers if any, too).
        if use_ssl:
            self.ssl_wouldblock_errors = tuple(frozenset((ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE)))
        self.socket_wouldblock_errors = tuple(frozenset((errno.EWOULDBLOCK, errno.EAGAIN)))
        if handler_class is None:
            self.handler = None
        else:
            self.handler = handler_class(*handler_class_args, **handler_class_kwargs)
        # The following looped code is perverse but saves more than 100 lines of "pythonic"
        # (and less DRY) code
        _routine_defaults = {
            'end_request': lambda _request: '\n' in _request,
            'split_request': lambda _request: (
              [_splitpoint for _splitpoint in (_request.find('\n') + 1,)],
              [_trailing for _trailing in (_splitpoint - 1 < len(_request),)],
              ((_request,''),(_request[:_splitpoint],_request[_splitpoint:]))[_trailing],
            )[-1],
            'handle_request': lambda _request: (_request, False),
            # See note at top of script. Now not doing "decoding" from utf8 for that reason.
            'decode_request': lambda _request: _request,
            'encode_response': lambda _response: _response, # Remember _response is: tuple(data, keepalive_flag)
        }
        for _routine in 'end_request', 'split_request', 'handle_request', 'decode_request', 'encode_response':
            if _proc.has_key(_routine):
                if isinstance(_proc[_routine], basestring):
                    _proc.update({_routine + '_name': str(_proc[_routine])})
                    if self.handler and hasattr(self.handler, _proc[_routine + '_name']) and inspect.method(getattr(self.handler, _proc[_routine + '_name'])):
                        _proc.update({_routine: getattr(self.handler, _proc[_routine + '_name'])})
                    elif self._globals.has_key(_proc[_routine + '_name']) and inspect.isroutine(self._globals[_proc[_routine + '_name']]):
                        _proc.update({_routine: self._globals[_proc[_routine + '_name']]})
                    elif _routine == _proc[_routine + '_name']:
                        _proc.update({_routine: _routine_defaults[_routine]})
                    else:
                        try:
                            if len(_proc[_routine]) > 7 and _proc[_routine][:7] == 'lambda ':
                                _proc.update({_routine: eval(_proc[_routine + '_name'])})
                            else:
                                raise SyntaxError
                        except SyntaxError:
                            raise ValueError('{} is not the name of a valid method or function, and doesn\'t look like a valid lambda. It is: {}'.format(_routine, _proc[_routine + '_name']))
                elif inspect.isroutine(_proc[_routine]):
                    if _proc[_routine + '_name'] in (None, ''):
                        _proc.update({_routine + '_name': str(_proc[_routine].__name__)})
                else:
                    raise TypeError('{} is not a valid function or method. It is: {}'.format(_routine, _proc[_routine]))
        for _prockey in _proc:
            setattr(self, _prockey, _proc[_prockey])

    def _setup(self):
        ## setup socket - ip version(s), host, port, backlog, ssl_opts
        if self.ip_version == 4:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            if self.ip_version == 6:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
            else:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.host, self.port))
        sock.listen(self.backlog)
        sock.setblocking(0)
        sock_fileno = sock.fileno()
        if self.use_ssl:
            self.ssl_opts = {'server_side': True, 'certfile': os.path.realpath(self.ssl_certfile), 'cert_reqs': self.ssl_cert_reqs, 'ca_certs': os.path.realpath(self.ssl_ca_certs), 'ssl_version': self.ssl_version}
            if self.ssl_keyfile is not None:
                self.ssl_opts.update({'keyfile': os.path.realpath(self.ssl_keyfile)})
        ## setup poller
        epoll = select.epoll()
        epoll.register(sock_fileno, select.EPOLLIN | select.EPOLLET)
        return sock, sock_fileno, epoll

    def ioloop(self, one_req=False):
        ## initial state
        finished_accepting = False
        shutdown_wanted = False
        response  = ''
        connections = {}
        requests = {}
        dec_requests = {}
        responses = {}
        keepalive_flags = {}
        sock, sock_fileno, epoll = self._setup()
        ## link directly to remaining self-vars from locals to avoid lookups and speedup the loop
        use_ssl = self.use_ssl
        poll_timeout = self.poll_timeout
        recv_size = self.recv_size
        verbose = self.verbose
        end_request = self.end_request
        handle_request = self.handle_request
        split_request = self.split_request
        decode_request = self.decode_request
        encode_response = self.encode_response
        e_register = epoll.register
        e_unregister = epoll.unregister
        e_modify = epoll.modify
        e_poll = epoll.poll
        e_close = epoll.close
        s_accept = sock.accept
        s_close = sock.close
        sel_EPOLLIN = select.EPOLLIN
        sel_EPOLLOUT = select.EPOLLOUT
        sel_EPOLLET = select.EPOLLET
        sel_EPOLLHUP = select.EPOLLHUP
        s_error = socket.error
        s_wouldblock_errors = self.socket_wouldblock_errors
        if use_ssl:
            ssl_opts = self.ssl_opts
            ssl_wrap_socket = ssl.wrap_socket
            ssl_error = ssl.SSLError
            _wouldblock_errors = self.ssl_wouldblock_errors
        else:
            _wouldblock_errors = s_wouldblock_errors
        die_on_error = self.die_on_error
        if die_on_error:
            if use_ssl:
                _sock_error = ssl_error
            else:
                _sock_error = s_error
        else:
            _sock_error = Exception
        ## this whole loop is not very DRY because I have "unrolled" it a bit, on purpose, for speed
        try:
            ## main polling loop
            while not shutdown_wanted:
                events = e_poll(poll_timeout)
                while events:
                    # Iterate through events list until it's empty, iterate backwards so removing items doesn't cause skips
                    for indexno in xrange(len(events) - 1, -1, -1):
                        fileno, event = events[indexno]
                        if fileno == sock_fileno:
                            ## accept new connection
                            if not finished_accepting:
                                try:
                                    connection, address = s_accept()
                                    if use_ssl:
                                        connection = ssl_wrap_socket(connection, **ssl_opts)
                                    connection.setblocking(0)
                                    connection_fileno = connection.fileno()
                                    e_register(connection_fileno, sel_EPOLLIN | sel_EPOLLET)
                                except (socket.error, ssl.SSLError) as e: # catch both types here
                                    if die_on_error:
                                        raise
                                    if one_req:
                                        shutdown_wanted = True
                                else:
                                    if one_req:
                                        finished_accepting = True
                                    connections[connection_fileno] = connection
                                    requests[connection_fileno] = ''
                                    dec_requests[connection_fileno] = ''
                                    responses[connection_fileno] = response
                                    keepalive_flags[connection_fileno] = False
                                    if verbose:
                                        log('=' * 40, 'setup socket ' + str(connection_fileno))
                        else:
                            try:
                                if event & sel_EPOLLIN:
                                    try:
                                        ## read data
                                        requests[fileno] += connections[fileno].recv(recv_size) # don't decode yet (see explanation below)
                                    except _sock_error as e:
                                        if e.args[0] in _wouldblock_errors:
                                            pass
                                    ## process request
                                    try:
                                        dec_requests[fileno] += decode_request(requests[fileno])
                                        requests[fileno] = ''
                                    except ValueError:
                                        # Capture decoding exceptions in case decoding a partial chunk would bork (e.g. on a multibyte
                                        # character split across a read boundary) but a valid request is contained in the beginning of
                                        # the data. Hope that eventually there will be a non-erroring chunk of data to decode...
                                        pass
                                    else:
                                        if end_request(dec_requests[fileno]):
                                            processing_request, dec_requests[fileno] = split_request(dec_requests[fileno])
                                            responses[fileno], keepalive_flags[fileno] = encode_response(handle_request(processing_request))
                                            e_modify(fileno, sel_EPOLLOUT | sel_EPOLLET)
                                            if verbose:
                                                log('-' * 40,
                                                    'processed response for socket ' + str(fileno),
                                                    ' decoded request: ' + repr(processing_request),
                                                    ' encoded response: ' + repr(responses[fileno]),
                                                    ' decoded leftovers: ' + repr(dec_requests[fileno]),
                                                    ' other leftovers: ' + repr(requests[fileno]))
                                elif event & sel_EPOLLOUT:
                                    ## write data
                                    try:
                                        while len(responses[fileno]) != 0:
                                            byteswritten = connections[fileno].send(responses[fileno])
                                            responses[fileno] = responses[fileno][byteswritten:]
                                    except _sock_error as e:
                                        if e.args[0] in _wouldblock_errors:
                                            pass
                                    else:
                                        if keepalive_flags[fileno]:
                                            e_modify(fileno, sel_EPOLLIN | sel_EPOLLET)
                                        else:
                                            e_unregister(fileno)
                                            connections[fileno].shutdown(socket.SHUT_RDWR)
                                            del connections[fileno], requests[fileno], dec_requests[fileno], responses[fileno], keepalive_flags[fileno]
                                            if one_req:
                                                shutdown_wanted = True
                                        if verbose:
                                            log('-' * 40, 'response output to socket ' + str(fileno))
                                elif event & sel_EPOLLHUP:
                                    ## hangup
                                    if verbose:
                                        log('-' * 40, 'hanging up socket ' + str(fileno))
                                    e_unregister(fileno)
                                    connections[fileno].shutdown(socket.SHUT_RDWR)
                                    connections[fileno].close()
                                    del connections[fileno], requests[fileno], dec_requests[fileno], responses[fileno], keepalive_flags[fileno]
                                    if one_req:
                                        shutdown_wanted = True
                            except Exception as e:
                                if die_on_error:
                                    raise
                                e_unregister(fileno)
                                connections[fileno].shutdown(socket.SHUT_RDWR)
                                connections[fileno].close()
                                del connections[fileno], requests[fileno], dec_requests[fileno], responses[fileno], keepalive_flags[fileno]
                                if one_req:
                                    shutdown_wanted = True
                        del events[indexno]
        finally:
            ## cleanup leftover connections
            for fileno in connections.keys(): # eager evaluate
                e_unregister(fileno)
                connections[fileno].shutdown(socket.SHUT_RDWR)
                connections[fileno].close()
            ## unregister and close server
            e_unregister(sock_fileno)
            e_close()
            s_close()
    
    def handle(self):
        self.ioloop(one_req=True)

    def serve(self):
        self.ioloop()

if __name__ == '__main__':
    # If executed directly this runs as a minimal one-line echo server (with
    # only LF end-of-lines, listening on all interfaces, port 8080, IPv4 & IPv6,
    # with no decoding or encoding, with no keepalive or logging to console) in
    # a loop until interrupted (by keyboard or signal).
    server = Gaidaros()
    server.serve()



#/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, with_statement

'''Setup and build script for Gaidaros.'''

#
# See README.rst for more information.
#
# This setup.py and file hierarchy is from the setup-with-teeth project (v0.2.6):
#  https://github.com/rowanthorpe/setup-with-teeth
#  replace all XXXXXXX with your own text

# FIXME: Make updating files with paths part of the build pass, not post-install
# FIXME: Make macro expansion in "bin/XXXXXXX" file more explicit (at the moment is too intrusive)
# FIXME: Generate project name from directory (even in virtualenv, pip, etc)
# TODO: add logic for using more vars from projects[], e.g. package_data, etc
# TODO: add more hosttypes to templates section
# TODO: Recursive macro-expansion
# TODO: Multiple authors -> iterable macros?
# TODO: Add more license types, host types, repo types, etc
# TODO(_set_dict_from_file()): Use proper python parsing => str, arr, or dict. Not readlines-type hacks.

if __name__ != '__main__':
    raise RuntimeError('This script should only be run directly. It is not a library.')

import os, sys, shutil, tempfile, glob, sysconfig, re, fcntl, subprocess, importlib
from os.path import join as p_join, dirname as p_dirname, realpath as p_realpath, basename as p_basename, \
                    isdir as p_isdir, isfile as p_isfile, pardir as p_pardir, getmtime as p_getmtime, \
                    normpath as p_normpath, splitdrive as p_splitdrive
from distutils.dir_util import mkpath as d_mkpath
from distutils.command.install_data import install_data as d_install_data
from distutils.command.install_scripts import install_scripts as d_install_scripts
from distutils.core import setup
from distutils.dist import DistributionMetadata

## GLOBAL VARS SETUP ##

_newdirsep = p_realpath('.')
_dirsep = ''
while _newdirsep != _dirsep: # iterate to find '/' or the system equivalent
    _dirsep = _newdirsep
    _newdirsep = p_dirname(_dirsep)
_dirsep = p_splitdrive(_dirsep)[1]
del _newdirsep
_projectpath = p_realpath('.')
_configvars = sysconfig.get_config_vars()
_configpaths = sysconfig.get_paths()
if p_basename(_configpaths['data']) == 'usr': #GOTCHA: '[path]/usr', not only '/usr', to allow for virtualenvs...
    _configprefix = p_normpath(p_join(_configpaths['data'], p_pardir, 'etc')) # "[path]/usr" => "[path]/etc" ("[path]/usr/etc", FHS-friendly)
else:
    _configprefix = p_join(_configpaths['data'], 'etc') # "[path]/[something_else]" => "[path]/[something_else]/etc"
_dirsep, _projectpath, _configprefix = unicode(_dirsep), unicode(_projectpath), unicode(_configprefix)

## TO EDIT ##

project = {
    'description': 'Async server micro-framework for control freaks',
    'hosttype': 'github',
    'repotype': 'git',
    'username': 'rowanthorpe',
    'author': 'Rowan Thorpe',
    'author_email': 'rowan@rowanthorpe.com',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Hardware :: Symmetric Multi-processing',
        'Topic :: System :: Networking',
    ],
    'keywords': ["async", "tcp", "server"],
    'macros_to_replace': ['name', 'username', 'description', 'author', 'author_email', 'version'],
    'license': None,          #NB: "...", if needed (not in classifiers)
    'py_modules': None,       #NB: [...], if needed
    'maintainer': None,       #NB: "...", if needed
    'maintainer_email': None, #NB: "...", if needed
    'platforms': None,        #NB: [...], if needed
}

## MAY BE EDITED ##

#project['name'] = unicode(p_basename(_projectpath)) #FIXME: doesn't play nice with virtualenv
project['name'] = 'gaidaros'
if p_join(_projectpath, 'lib') not in sys.path:
    sys.path.insert(0, p_join(_projectpath, 'lib'))
project.update({
    'files_to_expand': (p_join('lib', project['name'] + '.py'), 'README.rst'),
    'version': getattr(importlib.import_module(project['name']), '__version__'),
})
del sys.path[0]

## TEMPLATES ##

if project['hosttype'] == 'github':
    project['url_template'] = 'https://github.com/@username@/@name@'
    project['download_url_template'] = 'https://github.com/@username@/@name@/tarball/@version@'
#if project['license'] == '[weird unknown license]':
#    project['license_text'] = 'The [weird unknown] License: [url]'

## FUNCTIONS ##

def _files_glob(path, globs, trim_prefix='', realpath=False):
    if not path:
        path = '.'
    if realpath:
        path = p_realpath(path)
        if trim_prefix:
            trim_prefix = p_realpath(trim_prefix)
    elif trim_prefix:
        trim_prefix = p_normpath(trim_prefix)
    for globlet in globs:
        globresults = glob.glob(p_normpath(p_join(path, globlet)))
        for globresult in globresults:
            if trim_prefix and len(trim_prefix + _dirsep) < len(globresult) and globresult[:len(trim_prefix + _dirsep)] == trim_prefix + _dirsep:
                result = globresult[len(trim_prefix + _dirsep):]
            else:
                result = globresult
            yield unicode(result)

def _files_glob_l(*args, **kws):
    return list(_files_glob(*args, **kws))

def _find_dirs(topdir='lib'):
    if p_isdir(topdir):
        for _dir in os.walk(topdir):
            if p_isfile(p_join(_dir[0], '__init__.py')):
                yield _dir[0]

def _pkg_mappings(*args, **kws):
    yield project['name'], 'lib'
    for _dir in _find_dirs(*args, **kws):
        if _dir != 'lib':
            yield project['name'] + '.' + _dir[len('lib' + _dirsep):].replace(_dirsep, '.'), _dir # remove leading "lib/" and do s:/:.:g

def _regex_sub_lines(file_path, *pat_subs):
    fh, abs_path = tempfile.mkstemp()
    fcntl.lockf(fh, fcntl.LOCK_EX)
    with open(file_path, 'r') as old_file:
        with open(abs_path, 'w') as new_file:
            fcntl.lockf(new_file, fcntl.LOCK_EX)
            for line in old_file.readlines():
                line = line.rstrip('\r\n')
                for pattern, subst in pat_subs:
                    line = re.sub(pattern, subst, line)
                new_file.write(line + '\n')
    fcntl.lockf(fh, fcntl.LOCK_UN)
    os.close(fh)
    os.chmod(abs_path,os.stat(file_path)[0])
    os.remove(file_path)
    shutil.move(abs_path, file_path)

def _read_file(file):
    with open(file) as f:
        val = f.read()
    return val

def _readlines_file_as_array(file):
    with open(file) as f:
        arr = [x.rstrip('\r\n') for x in f.readlines()]
    return arr

def _readlines_file_as_dict(file):
    return dict(x.split(' ', 1) for x in _readlines_file_as_array(file))

def _set_dict_from_file(file, _dict, val, as_type):
    if as_type == 'array':
        _dict[val] = _readlines_file_as_array(file)
    elif as_type == 'dict':
        _dict[val] = _readlines_file_as_dict(file)
    else:
        _dict[val] = _read_file(file)

## OVERRIDE CLASSES ##

class MyInstallScripts(d_install_scripts):
    def run(self):
        d_install_scripts.run(self)
        _regex_sub_lines(p_join(_configpaths['scripts'], project['name']),
          (r'@configfile@',
           '"' + p_join(_configprefix, project['name'], project['name'] + r'.conf"')))

class MyInstallData(d_install_data):
    def run(self):
        d_mkpath(p_join(_configprefix, project['name'])) #GOTCHA: needed for virtualenvs, @prefix@/etc sometimes doesn't exist (even as symlink)
        d_install_data.run(self)
        _regex_sub_lines(p_join(_configprefix, project['name'], project['name'] + '.conf'),
          (r'^( *)basedir *:.*$',
           r'\1basedir: ' + _configvars['base']),
          (r'^( *)root *=.*$',
           r'\1root = %(basedir)s'),
          (r'^( *)lib *=.*$',
           r'\1lib = ' + re.sub(r'^' + _configvars['base'], r'%(basedir)s', p_join(_configpaths['purelib'], project['name']))),
          (r'^( *)scripts *=.*$',
           r'\1scripts = ' + re.sub(r'^' + _configvars['base'], r'%(basedir)s', _configpaths['scripts'])),
          (r'^( *)run *=.*$',
           r'\1run = ' + re.sub(r'^' + _configvars['base'], r'%(basedir)s', p_join(_configvars['base'], 'run', project['name']))),
          (r'^( *)configs *=.*$',
           r'\1configs = ' + re.sub(r'^' + _configvars['base'], r'%(basedir)s', p_join(_configprefix, project['name']))),
          (r'^( *)docs *=.*$',
           r'\1docs = ' + re.sub(r'^' + _configvars['base'], r'%(basedir)s', p_join(_configvars['base'], 'share', 'doc', project['name']))))

## MAIN ##

if sys.argv[1] == 'macros':
    if len(sys.argv) > 2 and sys.argv[2]:
        project['version'] = sys.argv[2]
    else:
        #NB: default to bumping the version up by one patch level, e.g.: 0.1.2 -> 0.1.3
        #    to bump a minor or major level, specify the version string manually as second arg instead
        if project['repotype'] == 'git':
            if not p_isdir('.git'):
                sys.stderr.write("You tried to update version information but the repo files appear not to be in this directory.\n".format(project['repotype']))
                sys.exit(1)
            _projectversionstr = subprocess.Popen(['git describe --abbrev=0'], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip('\r?\n').split('.')
        else:
            sys.stderr.write("Repo type {} not implemented yet.\n".format(project['repotype']))
            sys.exit(1)
        project['version'] = '.'.join(_projectversionstr[:-1] + [unicode(int(_projectversionstr[-1]) + 1)])
    for file_to_expand in project['files_to_expand']:
        file_to_expand = p_join('.', file_to_expand)
        if p_isfile(file_to_expand + '.in'):
            if p_isfile(file_to_expand) and p_getmtime(file_to_expand) > p_getmtime(file_to_expand + '.in'):
                sys.stderr.write('WARNING: File to generate "{}" is already newer than input file "{}". "touch" the input file to override. Skipping.\n'.format(file_to_expand, file_to_expand + '.in'))
            else:
                with open(file_to_expand + '.in', 'r') as fh_in:
                    file_content = fh_in.read().decode('utf-8')
                for macro_to_replace in project['macros_to_replace']:
                    file_content = re.sub(r'@' + macro_to_replace + r'@', project[macro_to_replace], file_content)
                with open(file_to_expand, 'w') as fh_out:
                    fcntl.lockf(fh_out, fcntl.LOCK_EX)
                    fh_out.write(file_content.encode('utf-8'))
        else:
            sys.stderr.write('WARNING: Specified file to expand "{}" does not have an input file "{}". Skipping.\n'.format(file_to_expand, file_to_expand + '.in'))
else:
    METADATA = {
        'cmdclass': {"install_data": MyInstallData, "install_scripts": MyInstallScripts},
        'name': str(project['name']), #GOTCHA: name must be a bytestring, even in unicode environment
        'version': project['version'],
        'provides': [project['name']],
        'description': project['description'],
        'author': project['author'],
        'author_email': project['author_email'],
        'scripts': _files_glob_l('bin', ('*',)),
        'data_files': [
            (p_join('share', 'doc', project['name']),
                _files_glob_l('.', ('*.rst', '*.txt'), trim_prefix='.') +
                _files_glob_l(p_join('doc','share'), ('*.rst', '*.txt', '*.pdf', '*.html'))),
            (p_join('share', 'doc', project['name'], 'examples'),
                _files_glob_l(p_join('doc', 'share', 'examples'), ('*.py', '*.sh'))),
            (p_join(_configprefix, project['name']),
                _files_glob_l(p_join('etc', project['name']), ('*.conf',))),
        ],
        'package_dir': dict(_pkg_mappings()),
        'package_data': {},
        'classifiers': project['classifiers'],
        'keywords': project['keywords'],
        'url': project['url_template'],
        'download_url': project['download_url_template'],
    }
    for _val in 'license', 'py_modules', 'maintainer', 'maintainer_email', 'platforms':
        if project[_val] is not None:
            METADATA[_val] = project[_val]
    for _man_section in xrange(1,9):
        _these_files = _files_glob_l(p_join('doc', 'man'), ('*.' + unicode(_man_section),))
        if _these_files:
            METADATA['data_files'] += (p_join('share', 'man', 'man' + unicode(_man_section)), _these_files)
    METADATA['packages'] = map(str, METADATA['package_dir'].keys())
    for _dir in METADATA['package_dir']:
        if p_isdir(p_join(METADATA['package_dir'][_dir], 'data')):
            _these_files = _files_glob_l(p_join(METADATA['package_dir'][_dir], 'data'), ('*.rst', '*.txt', '*.pdf', '*.html'))
            if _these_files:
                METADATA['package_data'][_dir] = _these_files
    for macro_to_expand in ['name', 'username', 'version', 'hosttype', 'repotype']:
        METADATA['url'] = re.sub(r'@' + macro_to_expand + r'@', project[macro_to_expand], METADATA['url'])
        METADATA['download_url'] = re.sub(r'@' + macro_to_expand + r'@', project[macro_to_expand], METADATA['download_url'])
    _set_dict_from_file('README.rst', METADATA, 'long_description', 'string')
    _set_dict_from_file('requirements.txt', METADATA, 'requires', 'array')

#NB: Don't need setuptools for now, using distutils only, but if used it would need something like:
#    try:
#        from setuptools import setup #, find_packages #NB: I do my own version of find_packages now
#        SETUPTOOLS_METADATA = {
#            'packages': find_packages(exclude=["*.test", "*.test.*", "test.*", "test"]),
#            'include_package_data': True,
#            'zip_safe': False,
#            'test_suite': '', #Not yet used
#            'entry_points': {}, #Not yet needed
#        }
#        _set_dict_from_file(....?, SETUPTOOLS_METADATA, 'install_requires', 'array')
#        _set_dict_from_file('requirements_extra.txt', SETUPTOOLS_METADATA, 'extras_require', 'dict')
#        METADATA.update(SETUPTOOLS_METADATA)
#    except ImportError:
#        if sys.version < '2.2.3':
#            DistributionMetadata.classifiers = None
#            DistributionMetadata.download_url = None

    if sys.version < '2.2.3':
        DistributionMetadata.classifiers = None
        DistributionMetadata.download_url = None

#DEBUG: begin
#    import pprint
#    pp = pprint.PrettyPrinter(indent=4)
#    pp.pprint(METADATA)
#    exit(0)
#DEBUG: end

    try:
        os.remove('MANIFEST') #GOTCHA: if this is present it doesn't get overwritten from present data
    except OSError:
        pass
    setup(**METADATA)

sys.exit(0)
